use super::*;
use core::sync::atomic::AtomicU64;
use futures_codec::{Bytes, BytesCodec, FramedRead, FramedWrite};
use futures_util::{
    stream::FuturesUnordered, AsyncReadExt, AsyncWriteExt, StreamExt, TryStreamExt,
};
use postcard::{from_bytes, to_stdvec};
use router_op_table::*;
use serde::*;
use std::io;
use stop_token::future::FutureExt as _;

struct RouterClientInner {
    jh_handler: Option<MustJoinHandle<()>>,
    stop_source: Option<StopSource>,
}

impl fmt::Debug for RouterClientInner {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_struct("RouterClientInner")
            .field("jh_handler", &self.jh_handler)
            .field("stop_source", &self.stop_source)
            .finish()
    }
}

struct RouterClientUnlockedInner {
    sender: flume::Sender<Bytes>,
    next_message_id: AtomicU64,
    router_op_waiter: RouterOpWaiter<ServerProcessorReplyResult, ()>,
}

impl fmt::Debug for RouterClientUnlockedInner {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_struct("RouterClientUnlockedInner")
            .field("sender", &self.sender)
            .field("next_message_id", &self.next_message_id)
            .field("router_op_waiter", &self.router_op_waiter)
            .finish()
    }
}

pub type MessageId = u64;
pub type SocketId = u64;

#[derive(Debug, PartialEq, Eq, Clone, Serialize, Deserialize)]
enum ServerProcessorRequest {
    AllocateMachine,
    ReleaseMachine {
        machine_id: MachineId,
    },
    GetInterfaces {
        machine_id: MachineId,
    },
    TcpConnect {
        machine_id: MachineId,
        local_address: Option<SocketAddr>,
        remote_address: SocketAddr,
        timeout_ms: u32,
        options: VirtualTcpOptions,
    },
    TcpBind {
        machine_id: MachineId,
        local_address: Option<SocketAddr>,
        options: VirtualTcpOptions,
    },
    TcpAccept {
        machine_id: MachineId,
        listen_socket_id: SocketId,
    },
    TcpShutdown {
        machine_id: MachineId,
        socket_id: SocketId,
    },
    UdpBind {
        machine_id: MachineId,
        local_address: Option<SocketAddr>,
        options: VirtualUdpOptions,
    },
    Send {
        machine_id: MachineId,
        socket_id: SocketId,
        data: Vec<u8>,
    },
    SendTo {
        machine_id: MachineId,
        socket_id: SocketId,
        remote_address: SocketAddr,
        data: Vec<u8>,
    },
    Recv {
        machine_id: MachineId,
        socket_id: u64,
        len: u32,
    },
    RecvFrom {
        machine_id: MachineId,
        socket_id: u64,
        len: u32,
    },
}

#[derive(Debug, PartialEq, Eq, Clone, Serialize, Deserialize)]
struct ServerProcessorMessage {
    message_id: MessageId,
    request: ServerProcessorRequest,
}

#[derive(Debug, PartialEq, Eq, Clone, Serialize, Deserialize)]
enum ServerProcessorCommand {
    Message(ServerProcessorMessage),
    CloseSocket {
        machine_id: MachineId,
        socket_id: SocketId,
    },
}

#[derive(Debug, PartialEq, Eq, Clone, Serialize, Deserialize)]
enum ServerProcessorReplyValue {
    AllocateMachine {
        machine_id: MachineId,
    },
    ReleaseMachine,
    GetInterfaces {
        interfaces: BTreeMap<String, NetworkInterface>,
    },
    TcpConnect {
        socket_id: SocketId,
        local_address: SocketAddr,
    },
    TcpBind {
        socket_id: SocketId,
        local_address: SocketAddr,
    },
    TcpAccept {
        socket_id: SocketId,
        address: SocketAddr,
    },
    TcpShutdown,
    UdpBind {
        socket_id: SocketId,
        local_address: SocketAddr,
    },
    Send {
        len: u32,
    },
    SendTo {
        len: u32,
    },
    Recv {
        data: Vec<u8>,
    },
    RecvFrom {
        remote_address: SocketAddr,
        data: Vec<u8>,
    },
}

#[derive(Debug, PartialEq, Eq, Clone, Serialize, Deserialize)]
enum ServerProcessorReplyResult {
    Value(ServerProcessorReplyValue),
    InvalidMachineId,
    InvalidSocketId,
    IoError(#[serde(with = "serde_io_error::SerdeIoErrorKindDef")] io::ErrorKind),
}

#[derive(Debug, PartialEq, Eq, Clone, Serialize, Deserialize)]
struct ServerProcessorReply {
    message_id: MessageId,
    status: ServerProcessorReplyResult,
}

#[derive(Debug, PartialEq, Eq, Clone, Serialize, Deserialize)]
enum ServerProcessorEvent {
    Reply(ServerProcessorReply),
    // DeadSocket {
    //     machine_id: MachineId,
    //     socket_id: SocketId,
    // },
}

#[derive(Debug, Clone)]
pub struct RouterClient {
    unlocked_inner: Arc<RouterClientUnlockedInner>,
    inner: Arc<Mutex<RouterClientInner>>,
}

impl RouterClient {
    //////////////////////////////////////////////////////////////////////////
    // Public interface

    #[cfg(not(all(target_arch = "wasm32", target_os = "unknown")))]
    pub async fn router_connect_tcp<H: ToSocketAddrs>(host: H) -> io::Result<RouterClient> {
        let addrs = host.to_socket_addrs()?.collect::<Vec<_>>();

        // Connect to RouterServer
        let ts_reader;
        let ts_writer;
        cfg_if! {
            if #[cfg(feature="rt-tokio")] {
                let ts = ::tokio::net::TcpStream::connect(addrs.as_slice()).await?;
                let (reader, writer) = ts.into_split();
                use tokio_util::compat::{TokioAsyncReadCompatExt, TokioAsyncWriteCompatExt};
                ts_reader = reader.compat();
                ts_writer = writer.compat_write();
            } else if #[cfg(feature="rt-async-std")] {
                use futures_util::io::AsyncReadExt;
                let ts = ::async_std::net::TcpStream::connect(addrs.as_slice()).await?;
                (ts_reader, ts_writer) = ts.split();
            } else {
                compile_error!("must choose an executor");
            }
        }

        // Create channels
        let (client_sender, server_receiver) = flume::unbounded::<Bytes>();

        // Create stopper
        let stop_source = StopSource::new();

        // Create router operation waiter
        let router_op_waiter = RouterOpWaiter::new();

        // Spawn a client connection handler
        let jh_handler = spawn(
            "RouterClient server processor",
            Self::run_server_processor(
                ts_reader,
                ts_writer,
                server_receiver,
                router_op_waiter.clone(),
                stop_source.token(),
            ),
        );

        Ok(Self::new(
            client_sender,
            router_op_waiter,
            jh_handler,
            stop_source,
        ))
    }

    #[cfg(all(target_arch = "wasm32", target_os = "unknown"))]
    pub async fn router_connect_ws<R: AsRef<str>>(request: R) -> io::Result<RouterClient> {
        let request = request.as_ref();

        // Connect to RouterServer
        let wsio_reader;
        let wsio_writer;
        cfg_if! {
            if #[cfg(feature="rt-wasm-bindgen")] {
                use ws_stream_wasm::*;
                let (_wsmeta, wsio) = WsMeta::connect(request, None)
                    .await.map_err(ws_err_to_io_error)?;
                use futures_util::io::AsyncReadExt;
                (wsio_reader, wsio_writer) = wsio.into_io().split();
            } else {
                compile_error!("must choose an executor");
            }
        }

        // Create channels
        let (client_sender, server_receiver) = flume::unbounded::<Bytes>();

        // Create stopper
        let stop_source = StopSource::new();

        // Create router operation waiter
        let router_op_waiter = RouterOpWaiter::new();

        // Spawn a client connection handler
        let jh_handler = spawn(
            "RouterClient server processor",
            Self::run_server_processor(
                wsio_reader,
                wsio_writer,
                server_receiver,
                router_op_waiter.clone(),
                stop_source.token(),
            ),
        );

        Ok(Self::new(
            client_sender,
            router_op_waiter,
            jh_handler,
            stop_source,
        ))
    }

    pub async fn disconnect(self) {
        drop(self.inner.lock().stop_source.take());
        let jh_handler = self.inner.lock().jh_handler.take();
        if let Some(jh_handler) = jh_handler {
            jh_handler.await;
        }
    }

    pub async fn allocate_machine(self) -> VirtualNetworkResult<MachineId> {
        let request = ServerProcessorRequest::AllocateMachine;
        let ServerProcessorReplyValue::AllocateMachine { machine_id } =
            self.perform_request(request).await?
        else {
            return Err(VirtualNetworkError::ResponseMismatch);
        };
        Ok(machine_id)
    }

    pub async fn release_machine(self, machine_id: MachineId) -> VirtualNetworkResult<()> {
        let request = ServerProcessorRequest::ReleaseMachine { machine_id };
        let ServerProcessorReplyValue::ReleaseMachine = self.perform_request(request).await? else {
            return Err(VirtualNetworkError::ResponseMismatch);
        };
        Ok(())
    }

    pub async fn get_interfaces(
        self,
        machine_id: MachineId,
    ) -> VirtualNetworkResult<BTreeMap<String, NetworkInterface>> {
        let request = ServerProcessorRequest::GetInterfaces { machine_id };
        let ServerProcessorReplyValue::GetInterfaces { interfaces } =
            self.perform_request(request).await?
        else {
            return Err(VirtualNetworkError::ResponseMismatch);
        };
        Ok(interfaces)
    }

    pub async fn tcp_connect(
        self,
        machine_id: MachineId,
        remote_address: SocketAddr,
        opt_local_address: Option<SocketAddr>,
        timeout_ms: u32,
        options: VirtualTcpOptions,
    ) -> VirtualNetworkResult<(SocketId, SocketAddr)> {
        let request = ServerProcessorRequest::TcpConnect {
            machine_id,
            local_address: opt_local_address,
            remote_address,
            timeout_ms,
            options,
        };
        let ServerProcessorReplyValue::TcpConnect {
            socket_id,
            local_address,
        } = self.perform_request(request).await?
        else {
            return Err(VirtualNetworkError::ResponseMismatch);
        };
        Ok((socket_id, local_address))
    }

    pub async fn tcp_bind(
        self,
        machine_id: MachineId,
        opt_local_address: Option<SocketAddr>,
        options: VirtualTcpOptions,
    ) -> VirtualNetworkResult<(SocketId, SocketAddr)> {
        let request = ServerProcessorRequest::TcpBind {
            machine_id,
            local_address: opt_local_address,
            options,
        };
        let ServerProcessorReplyValue::TcpBind {
            socket_id,
            local_address,
        } = self.perform_request(request).await?
        else {
            return Err(VirtualNetworkError::ResponseMismatch);
        };
        Ok((socket_id, local_address))
    }

    pub async fn tcp_accept(
        self,
        machine_id: MachineId,
        listen_socket_id: SocketId,
    ) -> VirtualNetworkResult<(SocketId, SocketAddr)> {
        let request = ServerProcessorRequest::TcpAccept {
            machine_id,
            listen_socket_id,
        };
        let ServerProcessorReplyValue::TcpAccept { socket_id, address } =
            self.perform_request(request).await?
        else {
            return Err(VirtualNetworkError::ResponseMismatch);
        };
        Ok((socket_id, address))
    }

    pub async fn tcp_shutdown(
        self,
        machine_id: MachineId,
        socket_id: SocketId,
    ) -> VirtualNetworkResult<()> {
        let request = ServerProcessorRequest::TcpShutdown {
            machine_id,
            socket_id,
        };
        let ServerProcessorReplyValue::TcpShutdown = self.perform_request(request).await? else {
            return Err(VirtualNetworkError::ResponseMismatch);
        };
        Ok(())
    }

    pub async fn udp_bind(
        self,
        machine_id: MachineId,
        opt_local_address: Option<SocketAddr>,
        options: VirtualUdpOptions,
    ) -> VirtualNetworkResult<(SocketId, SocketAddr)> {
        let request = ServerProcessorRequest::UdpBind {
            machine_id,
            local_address: opt_local_address,
            options,
        };
        let ServerProcessorReplyValue::UdpBind {
            socket_id,
            local_address,
        } = self.perform_request(request).await?
        else {
            return Err(VirtualNetworkError::ResponseMismatch);
        };
        Ok((socket_id, local_address))
    }

    pub async fn send(
        self,
        machine_id: MachineId,
        socket_id: SocketId,
        data: Vec<u8>,
    ) -> VirtualNetworkResult<usize> {
        let request = ServerProcessorRequest::Send {
            machine_id,
            socket_id,
            data,
        };
        let ServerProcessorReplyValue::Send { len } = self.perform_request(request).await? else {
            return Err(VirtualNetworkError::ResponseMismatch);
        };
        Ok(len as usize)
    }

    pub async fn send_to(
        self,
        machine_id: MachineId,
        socket_id: SocketId,
        remote_address: SocketAddr,
        data: Vec<u8>,
    ) -> VirtualNetworkResult<usize> {
        let request = ServerProcessorRequest::SendTo {
            machine_id,
            socket_id,
            data,
            remote_address,
        };
        let ServerProcessorReplyValue::SendTo { len } = self.perform_request(request).await? else {
            return Err(VirtualNetworkError::ResponseMismatch);
        };
        Ok(len as usize)
    }

    pub async fn recv(
        self,
        machine_id: MachineId,
        socket_id: u64,
        len: usize,
    ) -> VirtualNetworkResult<Vec<u8>> {
        let request = ServerProcessorRequest::Recv {
            machine_id,
            socket_id,
            len: len as u32,
        };
        let ServerProcessorReplyValue::Recv { data } = self.perform_request(request).await? else {
            return Err(VirtualNetworkError::ResponseMismatch);
        };
        Ok(data)
    }

    pub async fn recv_from(
        self,
        machine_id: MachineId,
        socket_id: u64,
        len: usize,
    ) -> VirtualNetworkResult<(Vec<u8>, SocketAddr)> {
        let request = ServerProcessorRequest::RecvFrom {
            machine_id,
            socket_id,
            len: len as u32,
        };
        let ServerProcessorReplyValue::RecvFrom {
            data,
            remote_address,
        } = self.perform_request(request).await?
        else {
            return Err(VirtualNetworkError::ResponseMismatch);
        };
        Ok((data, remote_address))
    }

    //////////////////////////////////////////////////////////////////////////
    // Private implementation

    fn new(
        sender: flume::Sender<Bytes>,
        router_op_waiter: RouterOpWaiter<ServerProcessorReplyResult, ()>,
        jh_handler: MustJoinHandle<()>,
        stop_source: StopSource,
    ) -> RouterClient {
        RouterClient {
            unlocked_inner: Arc::new(RouterClientUnlockedInner {
                sender,
                next_message_id: AtomicU64::new(0),
                router_op_waiter,
            }),
            inner: Arc::new(Mutex::new(RouterClientInner {
                jh_handler: Some(jh_handler),
                stop_source: Some(stop_source),
            })),
        }
    }

    fn report_closed_socket(&self, machine_id: MachineId, socket_id: SocketId) {
        let command = ServerProcessorCommand::CloseSocket {
            machine_id,
            socket_id,
        };
        let command_vec = match to_stdvec(&command).map_err(VirtualNetworkError::SerializationError)
        {
            Ok(v) => Bytes::from(v),
            Err(e) => {
                error!("{}", e);
                return;
            }
        };

        if let Err(e) = self
            .unlocked_inner
            .sender
            .send(command_vec)
            .map_err(|_| VirtualNetworkError::IoError(io::ErrorKind::BrokenPipe))
        {
            error!("{}", e);
        }
    }

    pub(super) fn drop_tcp_stream(&self, machine_id: MachineId, socket_id: SocketId) {
        self.report_closed_socket(machine_id, socket_id);
    }

    pub(super) fn drop_tcp_listener(&self, machine_id: MachineId, socket_id: SocketId) {
        self.report_closed_socket(machine_id, socket_id);
    }

    pub(super) fn drop_udp_socket(&self, machine_id: MachineId, socket_id: SocketId) {
        self.report_closed_socket(machine_id, socket_id);
    }

    async fn perform_request(
        &self,
        request: ServerProcessorRequest,
    ) -> VirtualNetworkResult<ServerProcessorReplyValue> {
        let message_id = self
            .unlocked_inner
            .next_message_id
            .fetch_add(1, Ordering::AcqRel);
        let command = ServerProcessorCommand::Message(ServerProcessorMessage {
            message_id,
            request,
        });
        let command_vec =
            Bytes::from(to_stdvec(&command).map_err(VirtualNetworkError::SerializationError)?);

        self.unlocked_inner
            .sender
            .send_async(command_vec)
            .await
            .map_err(|_| VirtualNetworkError::IoError(io::ErrorKind::BrokenPipe))?;
        let handle = self
            .unlocked_inner
            .router_op_waiter
            .add_op_waiter(message_id, ());

        let status = self
            .unlocked_inner
            .router_op_waiter
            .wait_for_op(handle)
            .await
            .map_err(|_| VirtualNetworkError::WaitError)?;

        match status {
            ServerProcessorReplyResult::Value(server_processor_response) => {
                Ok(server_processor_response)
            }
            ServerProcessorReplyResult::InvalidMachineId => {
                Err(VirtualNetworkError::InvalidMachineId)
            }
            ServerProcessorReplyResult::InvalidSocketId => {
                Err(VirtualNetworkError::InvalidSocketId)
            }
            ServerProcessorReplyResult::IoError(k) => Err(VirtualNetworkError::IoError(k)),
        }
    }

    async fn run_server_processor<R, W>(
        reader: R,
        writer: W,
        receiver: flume::Receiver<Bytes>,
        router_op_waiter: RouterOpWaiter<ServerProcessorReplyResult, ()>,
        stop_token: StopToken,
    ) where
        R: AsyncReadExt + Unpin + Send,
        W: AsyncWriteExt + Unpin + Send,
    {
        let mut unord = FuturesUnordered::new();

        let framed_reader = FramedRead::new(reader, BytesCodec);
        let framed_writer = FramedWrite::new(writer, BytesCodec);

        let framed_writer_fut = system_boxed(async move {
            if let Err(e) = receiver.into_stream().map(Ok).forward(framed_writer).await {
                error!("{}", e);
            }
        });
        let framed_reader_fut = system_boxed(async move {
            let fut = framed_reader.try_for_each(|x| async {
                let x = x;
                let evt = from_bytes::<ServerProcessorEvent>(&x)
                    .map_err(VirtualNetworkError::SerializationError)?;

                match evt {
                    ServerProcessorEvent::Reply(reply) => {
                        router_op_waiter
                            .complete_op_waiter(reply.message_id, reply.status)
                            .map_err(io::Error::other)?;
                    } // ServerProcessorEvent::DeadSocket {
                      //     machine_id,
                      //     socket_id,
                      // } => {
                      //     //
                      // }
                }

                Ok(())
            });
            if let Err(e) = fut.await {
                error!("{}", e);
            }
        });

        unord.push(framed_writer_fut);
        unord.push(framed_reader_fut);
        while let Ok(Some(_)) = unord.next().timeout_at(stop_token.clone()).await {}
    }
}
